<?php

namespace App\Command;

use App\Constants\AppConstants;
use Symfony\Component\Console\Attribute\AsCommand;
use Symfony\Component\Console\Command\Command;
use Symfony\Component\Console\Input\InputInterface;
use Symfony\Component\Console\Output\OutputInterface;
use Symfony\Component\Console\Style\SymfonyStyle;
use Symfony\Component\DependencyInjection\ContainerInterface;

use App\Entity\ModelMetadata\ModelMetadata;
use Symfony\Component\Serializer\Encoder\JsonEncoder;
use Symfony\Component\Serializer\Normalizer\ObjectNormalizer;
use Symfony\Component\Serializer\Serializer;

#[AsCommand(
    name: 'app:language:support',
    description: 'Generate a JSON file for dynamic modules with default english language settings to serve as a reference for translation.',
)]
class LanguageSupportCommand extends Command
{
    public function __construct(protected ContainerInterface $container)
    {
        $this->entityManager = $this->container->get('doctrine.orm.entity_manager');
        parent::__construct();
    }

    public const LOCALES_PATH = AppConstants::LOCALES_MMD_PATH;
    public const EN_JSON = 'en.json';
    public const MODULE_DETAILS = 'MODULE_DETAILS';
    public const SINGULAR_NAME = 'SINGULAR_NAME';
    public const PLURAL_NAME = 'PLURAL_NAME';
    public const FIELDS = 'FIELDS';

    protected function configure(): void
    {
        $this
            ->setName('app:language:support')
            ->setDescription('Generate a JSON file for dynamic modules with default english language settings to serve as a reference for translation.')
        ;
    }

    protected function execute(InputInterface $input, OutputInterface $output): int
    {
        $io = new SymfonyStyle($input, $output);

        $modelMetadatasWithKeys = $this->getAllMetadata(ModelMetadata::class);
        $totalNumberOfFilesToBeCreated = count($modelMetadatasWithKeys);
        $numberOfFilesSuccessfullyCreated = 0;
        $filesNotAbleToCreate = [];

        $encoder = new JsonEncoder();
        $normalizer = new ObjectNormalizer();
        $defaultContext = [];
        $defaultContext['circular_reference_handler'] = fn($modelMetadatasWithKeys) => $modelMetadatasWithKeys->getUuid();

        $serializer = new Serializer([$normalizer], [$encoder]);
        $normalizedModelMetadata = $serializer->serialize($modelMetadatasWithKeys, 'json', $defaultContext);
        $normalizedModelMetadata = json_decode($normalizedModelMetadata, true);

        foreach ($normalizedModelMetadata as $modelType => $modelMetadata) {
            $modelMapping = [];
            $singularName = $modelMetadata["descriptions"]["singular"];
            $pluralName = $modelMetadata["descriptions"]["plural"];
            $moduleDetails = array(self::SINGULAR_NAME => $singularName, self::PLURAL_NAME => $pluralName);
            $modelTypeUpperCase = strtoupper($modelType);
            $modelMapping[$modelTypeUpperCase] = array(self::MODULE_DETAILS => $moduleDetails);
            $modelAttributes = $modelMetadata['attributes'];
            $fieldsSummary = [];
            foreach ($modelAttributes as $attribute) {
                $fieldName = strtoupper($attribute['name']);
                $fieldTitle = $attribute['descriptions']['singular'];
                $fieldsSummary[$fieldName] = $fieldTitle;
            }
            $modelMapping[$modelTypeUpperCase][self::FIELDS] = $fieldsSummary;

            if (!\is_dir(self::LOCALES_PATH)) {
                mkdir(self::LOCALES_PATH, 0755, false);
            }
            $modelFolder = self::LOCALES_PATH . "/" . $modelType;
            if (!\is_dir($modelFolder)) {
                mkdir($modelFolder, 0755, true);
            }
            $filePath = sprintf("%s/%s", $modelFolder, self::EN_JSON);
            $result = file_put_contents($filePath, json_encode($modelMapping, JSON_PRETTY_PRINT));

            if ($result === false) {
                array_push($filesNotAbleToCreate, "/" . $modelType . "/" . self::EN_JSON);
            } else {
                $numberOfFilesSuccessfullyCreated = $numberOfFilesSuccessfullyCreated + 1;
            }
        }
        $msg = sprintf("Successfully created files at %s", self::LOCALES_PATH);
        if ($numberOfFilesSuccessfullyCreated !== $totalNumberOfFilesToBeCreated) {
            $msg = sprintf("%u out of %u files got created at %s", $numberOfFilesSuccessfullyCreated, $totalNumberOfFilesToBeCreated, self::LOCALES_PATH);
            $io->writeln("The following files were not created:");
            $io->listing($filesNotAbleToCreate);
            $io->error($msg);
            return Command::FAILURE;
        }
        $io->success($msg);
        return Command::SUCCESS;
    }

    /**
     * @param $class
     * @return array
     */
    private function getAllMetadata($class)
    {
        $metadataDocumentManager = $this->container->get('doctrine.orm.entity_manager');

        /** @var EntityRepository $modelMetadataRepository */
        $modelMetadataRepository = $metadataDocumentManager->getRepository($class);

        /** @var ModelMetadata[] $modelMetadatas */
        $modelMetadatas = $modelMetadataRepository->findAll();
        $modelMetadatasWithKeys = [];

        foreach ($modelMetadatas as $modelMetadata) {
            $modelMetadatasWithKeys[$modelMetadata->getType()] = $modelMetadata;
        }
        return $modelMetadatasWithKeys;
    }
}