<?php

namespace App\EventSubscriber;

use ApiPlatform\Core\Api\IriConverterInterface;
use ApiPlatform\Core\EventListener\EventPriorities;
use App\Constants\AppConstants;
use App\Constants\EntityFieldConstants;
use App\Constants\ErrorConstants;
use App\Service\MetadataHelper;
use App\Service\PreprocessingRuleService;
use App\Service\UtilityService;
use Doctrine\Inflector\InflectorFactory;
use Psr\Log\LoggerInterface;
use Symfony\Component\DependencyInjection\ContainerInterface;
use Symfony\Component\EventDispatcher\EventSubscriberInterface;
use Symfony\Component\HttpFoundation\Request;
use Symfony\Component\HttpKernel\Event\ViewEvent;
use Symfony\Component\HttpKernel\KernelEvents;
use Symfony\Component\Serializer\SerializerInterface;

class EvaluatePostProcessingRules implements EventSubscriberInterface
{

    const MAIN_FUNC = 'handlePostProcessRule';


    private $inflector;

    public function __construct(
        private ContainerInterface       $container,
        private IriConverterInterface    $iriConverter,
        private LoggerInterface          $logger,
        private MetadataHelper           $metadataHelper,
        private SerializerInterface      $serializer,
        private UtilityService           $utilityService,
        private PreprocessingRuleService $ruleService
    ) {
        $this->inflector = InflectorFactory::create()->build();
    }

    public static function getSubscribedEvents(): array
    {
        return [
            KernelEvents::VIEW => [self::MAIN_FUNC, EventPriorities::PRE_WRITE]
        ];
    }

    /**
     * @throws \Exception
     */
    public function handlePostProcessRule(ViewEvent $event)
    {
        $request = $event->getRequest();
        $requestMethod = $request->getMethod();
        $subRequestType = $request->attributes->get(AppConstants::ATTR_FSR_SUB_REQUEST_TYPE);
        $prevData = $request->attributes->get(AppConstants::ATTR_FSR_PREVIOUS_DATA, null);

        if (
            $requestMethod != Request::METHOD_POST
            || $subRequestType == AppConstants::FSR_SUB_REQUEST_IMPORT
            || !empty($prevData)
        ) {
            return;
        }

        $object = $event->getControllerResult();
        if (!is_object($object)) {
            return;
        }
        $entityClass = $object::class;
        if ($this->ruleService->validateRecord($entityClass, $object)) {
            return;
        }

        $entityType = $this->metadataHelper->getTypeFromClassName($entityClass);
        $activeRules = $this->filterRules($entityType);
        if (empty($activeRules)) {
            return;
        }
        $finalRules = $this->ruleService->filterByRecordType($activeRules, AppConstants::RECORD_TYPE_EXISTING);
        if (!empty($finalRules) && $this->ruleService->isModuleTrackable($object)) {
            $this->processRule($finalRules, $entityType, $entityClass, $object, $request);
        } else {
            // Can not execute rule because query needed createDate for dates filtering
            $this->logger->error(sprintf(ErrorConstants::FSR_CH_0000019, $entityClass));
        }
    }


    /**
     * @param array $rules
     * @param string $entityType
     * @param string $entityClass
     * @param mixed $object
     * @param Request $request
     * @return bool
     */
    public function processRule(array $rules, string $entityType, string $entityClass, mixed $object, Request $request): bool
    {
        $record = json_decode($request->getContent(), true);
        $record['uuid'] = $object->getUuid();

        foreach ($rules as $rule) {
            $records = $this->execute($entityClass, $rule, $record);

            // if no records found continue with next "Rule" evaluation
            if (empty($records)) continue;

            $this->processMatchedRule($rule, $object, $entityClass, $records[0]);
            $this->recordLogging($record, $entityType, $rule);

            // Set Skip playbook execution
            $request->attributes->set(AppConstants::SKIP_PLAYBOOK_EXECUTION, $rule[EntityFieldConstants::RULE_ATTR_SKIP_PL_EXEC]);
            return true;
        }
        return false;
    }

    /**
     * @param $rule
     * @param $object
     * @param $entityClass
     * @param $linkToRecord
     * @return void
     */
    public function processMatchedRule($rule, $object, $entityClass, $linkToRecord): void
    {
        $actionType = $rule[EntityFieldConstants::RULE_ATTR_ACTION_TYPE];
        $action = $rule[EntityFieldConstants::RULE_ATTR_ACTION];
        $linkingAllowed = (strcasecmp(AppConstants::RULE_ACTION_UPDATE, $actionType) == 0) && ($action["link"] ?? false);

        $isSelfLink = $this->isModuleSelfLinked($entityClass);
        if (!$linkingAllowed || !$isSelfLink) $linkToRecord = null;

        $this->updateRecords($object, $action, $entityClass, $linkToRecord);
    }


    /**
     * @param $object
     * @param $actionInfo
     * @param $entityClass
     * @param $linkObject
     * @return void
     */
    private function updateRecords($object, $actionInfo, $entityClass, $linkObject): void
    {
        $moduleType = $this->metadataHelper->getTypeFromClassName($entityClass);
        $action = $actionInfo['update'] ?? [];
        $object = $this->serializer->deserialize(
            json_encode($action),
            $entityClass,
            'jsonld',
            ['object_to_populate' => $object, 'fetch_data' => false]
        );

        $this->linkRecord($linkObject, $moduleType, $object);
    }

    /**
     * @param $linkObject
     * @param mixed $moduleType
     * @param $object
     * @return void
     */
    public function linkRecord($linkObject, mixed $moduleType, $object): void
    {
        if (!is_null($linkObject)) {
            $camelize = $this->inflector->classify($moduleType);
            $singulars = $this->inflector->singularize($camelize);
            $addMethod = 'add' . $singulars;
            $object->{$addMethod}($linkObject);
        }
    }

    /**
     * @param string $entityClass
     * @return bool
     */
    private function isModuleSelfLinked(string $entityClass): bool
    {
        $moduleType = $this->metadataHelper->getTypeFromClassName($entityClass);
        $moduleFields = $this->metadataHelper->getEntityFields($entityClass) ?? [];

        if (array_key_exists(AppConstants::ENTITY_ATTR_MAP_RELATED_FIELDS, $moduleFields)) {
            $relatedFieldMap = $moduleFields[AppConstants::ENTITY_ATTR_MAP_RELATED_FIELDS];
            return in_array($moduleType, $relatedFieldMap);
        }

        return false;
    }

    /**
     * @param mixed $entityType
     * @return array
     * @throws \Exception
     */
    public function filterRules(mixed $entityType): array
    {
        $rules = $this->ruleService->getRules($entityType);
        if (empty($rules)) {
            return $rules;
        }

        // filter drop rules
        $rules = $this->ruleService->filterByRuleType($rules, AppConstants::RULE_ACTION_UPDATE);

        // filter active rules
        $rules = $this->ruleService->filterByEndDate($rules);

        return $rules;
    }

    /**
     * @param $data
     * @param $entityType
     * @return void
     */
    private function recordLogging($data, $entityType, $rule): void
    {
        $dataToBroadcast = $this->prepareBroadcastData($data, $entityType, $rule);
        $this->sendData([$dataToBroadcast]);
    }

    /**
     * /**
     * @param $record
     * @param $entityType
     * @param $rule
     * @return array
     */
    private function prepareBroadcastData($record, $entityType, $rule): array
    {
        $auditData = [];
        $auditData['data'] = $record;
        $auditData['rule'] = $rule;

        return [
            AppConstants::COMPONENT => AppConstants::CRUD_HUB,
            AppConstants::ENTITY_TYPE => $entityType,
            AppConstants::HTTP_METHOD => AppConstants::AUDIT_OPERATION_UPDATE,
            AppConstants::REMOTE_ADDR => $_SERVER['HTTP_X_FORWARDED_FOR'] ?? $_SERVER['REMOTE_ADDR'],
            AppConstants::TRANSACTION_DATE => (int)($_SERVER['REQUEST_TIME_FLOAT'] * 1000),
            AppConstants::DELTA_DATA => $auditData,
        ];
    }

    /**
     * @param array $data
     * @param string $key
     * @return void
     */
    public function sendData(array $data, string $key = AppConstants::DEFALUT_KEY): void
    {
        $this->container->get(AppConstants::AUDIT_PRODUCER)->setContentType(AppConstants::CONTENT_TYPE_JSON);
        $this->container->get(AppConstants::AUDIT_PRODUCER)->publish(json_encode($data), $key);
    }

    /**
     * @param string $entityClass
     * @param $rule
     * @param $record
     * @return float|int|mixed|string
     */
    public function execute(string $entityClass, $rule, $record): mixed
    {
        try {
            $criteria = $this->ruleService->prepareRuleCriteria($rule, $record, true);
            return  $this->ruleService->executeRuleQuery($entityClass, $criteria, false);
        } catch (\Exception $e) {
            $this->logger->error($e);
            $dataToBroadcast = $this->ruleService->prepareRuleExecutionFailedBroadcastData($rule);
            $this->sendData([$dataToBroadcast]);
            return [];
        }
    }
}
