<?php
/* Copyright start
  Copyright (C) 2008 - 2023 Fortinet Inc.
  All rights reserved.
  FORTINET CONFIDENTIAL & FORTINET PROPRIETARY SOURCE CODE
  Copyright end */

namespace App\EventSubscriber;


use ApiPlatform\Core\EventListener\EventPriorities;
use App\Constants\AppConstants;
use App\Constants\EntityFieldConstants;
use App\Constants\ErrorConstants;
use App\Entity\Rules\PreprocessingRule;
use App\Service\MetadataHelper;
use App\Service\PreprocessingRuleService;
use App\Utility\FilterUtility;
use Doctrine\ORM\Exception\NotSupported;
use Psr\Log\LoggerInterface;
use Symfony\Component\DependencyInjection\ContainerInterface;
use Symfony\Component\EventDispatcher\EventSubscriberInterface;
use Symfony\Component\HttpFoundation\JsonResponse;
use Symfony\Component\HttpFoundation\Request;
use Symfony\Component\HttpFoundation\Response;
use Symfony\Component\HttpKernel\Event\ViewEvent;
use Symfony\Component\HttpKernel\KernelEvents;

class EvaluatePreProcessingRules implements EventSubscriberInterface
{
    private const MAIN_FUNC = 'handlePreProcessRule';

    public function __construct(
        private ContainerInterface       $container,
        private FilterUtility            $filterUtility,
        private LoggerInterface          $logger,
        private MetadataHelper           $metadataHelper,
        private PreprocessingRuleService $ruleService
    ) {
    }

    public static function getSubscribedEvents(): array
    {
        return [
            KernelEvents::VIEW => [self::MAIN_FUNC, EventPriorities::POST_VALIDATE]
        ];
    }

    /**
     * @throws \Exception
     */
    public function handlePreProcessRule(ViewEvent $event)
    {
        $request = $event->getRequest();
        $requestMethod = $request->getMethod();
        $subRequestType = $request->attributes->get(AppConstants::ATTR_FSR_SUB_REQUEST_TYPE);
        $prevData = $request->attributes->get(AppConstants::ATTR_FSR_PREVIOUS_DATA, null);
        $replication = (bool) $request->get('$' . AppConstants::IS_REPLICATION_FLAG, false);

        if (
            $requestMethod != Request::METHOD_POST
            || $subRequestType == AppConstants::FSR_SUB_REQUEST_IMPORT
            || !empty($prevData)
            || $replication
        ) {
            return;
        }

        $object = $event->getControllerResult();
        if (!is_object($object)) {
            return;
        }
        $entityClass = $object::class;
        if ($this->ruleService->validateRecord($entityClass, $object)) {
            return;
        }
        $entityType = $this->metadataHelper->getTypeFromClassName($entityClass);
        $dropRules = $this->filterRules($entityType);
        if (empty($dropRules)) {
            return;
        }
        $record = json_decode($request->getContent(), true);
        $ruleMatched = $this->executeIncomingDataRule($dropRules, $record);
        if (!is_null($ruleMatched)) return $this->processMatchedRule($ruleMatched, $record, $entityType, $event);

        // if module is not trackable then do not execute rule
        $isTrackable = $this->ruleService->isModuleTrackable($object);
        if (!$isTrackable) {
            $this->logger->error(sprintf(ErrorConstants::FSR_CH_0000019, $entityClass));
            return;
        }
        $ruleMatched = $this->executeExitingDataRule($dropRules, $record, $entityClass);
        if (!is_null($ruleMatched)) return $this->processMatchedRule($ruleMatched, $record, $entityType, $event);
    }

    /**
     * @param array $dropRules
     * @param mixed $normalized_data
     * @return array
     */
    public function executeIncomingDataRule(array $dropRules, mixed $normalized_data): mixed
    {
        $incomingDataRules = $this->ruleService->filterByRecordType($dropRules, AppConstants::RECORD_TYPE_INCOMING);
        foreach ($incomingDataRules as $rule) {
            $criteria = $rule[EntityFieldConstants::RULE_ATTR_CRITERIA];
            $condition = $criteria['condition'];
            $validate = $this->filterUtility->evaluateObject($condition, $normalized_data, []);
            if (!$validate) continue;
            return $rule;
        }
        return null;
    }

    /**
     * @param array $dropRules
     * @param mixed $record
     * @param string $entityClass
     * @return mixed
     * @throws NotSupported
     */
    public function executeExitingDataRule(array $dropRules, mixed $record, string $entityClass): mixed
    {
        $rules = $this->ruleService->filterByRecordType($dropRules, AppConstants::RECORD_TYPE_EXISTING);
        if (empty($rules)) {
            return null;
        }
        foreach ($rules as $rule) {
            $recUuid = $this->execute($entityClass, $rule, $record);
            if (empty($recUuid)) continue;
            return $rule;
        }
        return null;
    }

    /**
     * @param mixed $entityType
     * @return array
     * @throws \Exception
     */
    public function filterRules(mixed $entityType): array
    {
        /**
         * @var PreprocessingRule[] $rules
         */
        $rules = $this->ruleService->getRules($entityType);
        if (empty($rules)) {
            return $rules;
        }

        // filter drop rules
        $rules = $this->ruleService->filterByRuleType($rules, AppConstants::RULE_ACTION_DROP);

        // filter active rules
        $rules = $this->ruleService->filterByEndDate($rules);

        // filter rules by entityClass
        return $rules;
    }

    /**
     * @param $rule
     * @param mixed $data
     * @param mixed $entityType
     * @param ViewEvent $event
     * @return JsonResponse
     */
    public function processMatchedRule($rule, mixed $data, mixed $entityType, ViewEvent $event): JsonResponse
    {
        $this->recordLogging($data, $entityType, $rule);
        return $this->prepareResponse($data, $event, $rule);
    }


    /**
     * @param mixed $data
     * @param ViewEvent $event
     * @return JsonResponse
     */
    public function prepareResponse(mixed $data, ViewEvent $event, array $rule): JsonResponse
    {
        $msg = ['message' => sprintf(ErrorConstants::FSR_CH_0000018, $rule['name'])];
        $response = new JsonResponse($msg, Response::HTTP_MULTI_STATUS);
        $event->setResponse($response);
        return $response;
    }

    /**
     * @param $data
     * @param $entityType
     * @return void
     */
    private function recordLogging($data, $entityType, $rule): void
    {
        $dataToBroadcast = $this->prepareBroadcastData($data, $entityType, $rule);
        $this->sendData([$dataToBroadcast]);
    }


    /**
     * @param $record
     * @param $entityType
     * @return array
     */
    private function prepareBroadcastData($record, $entityType, $rule): array
    {
        $auditData = [];
        $auditData['data'] = $record;
        $auditData['rule'] = $rule;

        return [
            AppConstants::COMPONENT => AppConstants::CRUD_HUB,
            AppConstants::ENTITY_TYPE => $entityType,
            AppConstants::HTTP_METHOD => AppConstants::AUDIT_OPERATION_DROP,
            AppConstants::REMOTE_ADDR => $_SERVER['HTTP_X_FORWARDED_FOR'] ?? $_SERVER['REMOTE_ADDR'],
            AppConstants::TRANSACTION_DATE => (int)($_SERVER['REQUEST_TIME_FLOAT'] * 1000),
            AppConstants::DELTA_DATA => $auditData,
        ];
    }

    /**
     * @param array $data
     * @param string $key
     * @return void
     */
    public function sendData(array $data, string $key = AppConstants::DEFALUT_KEY): void
    {
        $this->container->get(AppConstants::AUDIT_PRODUCER)->setContentType(AppConstants::CONTENT_TYPE_JSON);
        $this->container->get(AppConstants::AUDIT_PRODUCER)->publish(json_encode($data), $key);
    }

    /**
     * @param string $entityClass
     * @param $rule
     * @param $record
     * @return float|int|mixed|string
     */
    public function execute(string $entityClass, $rule, $record): mixed
    {
        try {
            $criteria = $this->ruleService->prepareRuleCriteria($rule, $record, false);
            return $this->ruleService->executeRuleQuery($entityClass, $criteria, true);
        } catch (\Exception $e) {
            $this->logger->error($e);
            $dataToBroadcast = $this->ruleService->prepareRuleExecutionFailedBroadcastData($rule);
            $this->sendData([$dataToBroadcast]);
            return [];
        }
    }
}
