<?php
/* Copyright start
  Copyright (C) 2008 - 2024 Fortinet Inc.
  All rights reserved.
  FORTINET CONFIDENTIAL & FORTINET PROPRIETARY SOURCE CODE
  Copyright end */

namespace App\Providers;


use stdClass;
use GuzzleHttp\Client;
use Psr\Log\LoggerInterface;
use App\Constants\AppConstants;
use App\Constants\CacheConstants;
use App\Constants\ErrorConstants;
use GuzzleHttp\Psr7\Request as GuzzleRequest;
use Symfony\Component\HttpFoundation\Request;
use Symfony\Contracts\Cache\TagAwareCacheInterface;
use Symfony\Component\DependencyInjection\ContainerInterface;
use Symfony\Component\HttpKernel\Exception\BadRequestHttpException;
use Symfony\Component\Security\Core\Exception\AuthenticationException;

class ApiKeyProvider
{

    protected $dasClient;

    public function __construct(
        private ContainerInterface $container,
        protected TagAwareCacheInterface $cache,
        protected LoggerInterface $logger

    ) {
        $this->dasClient = $this->container->get("app.proxy.client.das");
    }

    /**
     * @param $theirFingerprint
     * @param Request $request
     * @param int $timeout Seconds
     * @return string $userId;
     */
    public function validateFingerprint($theirFingerprint, Request $request = null)
    {
        if (!$request || $request->attributes->get('fsr_subrequest_type') !== null) {
            return true;
        }
        $protected_key = hash('sha256', $theirFingerprint);
        $item = $this->cache->getItem($protected_key);
        if ($item->isHit()) {
            return $item->get();
        }
        $body = new stdClass();
        $body->{'api_key'} = $theirFingerprint;
        $headers = array(
            "Content-type" => "application/json",
        );
        $method = 'POST';
        $uri = sprintf('%s/api-keys/validate', $this->dasClient->getTargetUri());
        $this->logger->info(" api key validate uri: " .  $uri);
        try {
            $guzzleRequest = new GuzzleRequest($method, $uri, $headers, json_encode($body));
            $response = $this->dasClient->send($guzzleRequest, ['verify' => false]);
        } catch (BadRequestHttpException | \Exception $error) {
            $this->logger->error($error->getMessage());
            throw new AuthenticationException(ErrorConstants::FSR_CH_0000015);
        }
        $decodedResponse = json_decode($response->getBody(), true);
        # The uuid in DAS, corresponds to user_id in actors table.
        $userId = $decodedResponse['uuid'];
        # The validUntil will always comes in epoch.
        $validUntil = $decodedResponse['valid_until'];
        $cacheTime = $this->identifyCacheTime($validUntil);
        $item->set($userId);
        $item->expiresAfter($cacheTime);
        $item->tag(CacheConstants::API_ACCESS_TOKEN);
        $this->cache->save($item);
        return $userId;
    }

    /**
     * @param int $validUntil
     * @return int $cacheTime;
     */
    private function identifyCacheTime(int $validUntil)
    {
        $this->logger->debug($validUntil);
        $diff = $validUntil - time();
        if ($diff > AppConstants::CACHE_EXPIRY_FOR_API_KEY) {
            return AppConstants::CACHE_EXPIRY_FOR_API_KEY;
        }
        return $diff;
    }
}
