<?php
/* Copyright start
  Copyright (C) 2008 - 2024 Fortinet Inc.
  All rights reserved.
  FORTINET CONFIDENTIAL & FORTINET PROPRIETARY SOURCE CODE
  Copyright end */

namespace App\Security;

use App\Entity\Actors\Actor;
use Psr\Log\LoggerInterface;
use App\Constants\ErrorConstants;
use App\Providers\ApiKeyProvider;
use App\Security\BaseAuthenticator;
use Doctrine\ORM\EntityManagerInterface;
use App\Security\HmacPostAuthenticationToken;
use Symfony\Component\HttpFoundation\Request;
use Symfony\Contracts\Cache\TagAwareCacheInterface;
use Symfony\Component\DependencyInjection\ContainerInterface;
use Symfony\Component\Security\Http\Authenticator\Passport\Passport;
use Symfony\Component\Security\Core\Authentication\Token\TokenInterface;
use Lexik\Bundle\JWTAuthenticationBundle\Services\JWTTokenManagerInterface;
use Symfony\Component\Security\Http\Authenticator\Passport\Badge\UserBadge;
use Symfony\Component\Security\Http\Authenticator\Passport\SelfValidatingPassport;
use Symfony\Component\Security\Core\Exception\CustomUserMessageAuthenticationException;
use Lexik\Bundle\JWTAuthenticationBundle\TokenExtractor\AuthorizationHeaderTokenExtractor;

class ApiKeyAuthenticator extends BaseAuthenticator
{
    protected $userRepository;

    public function __construct(
        protected JWTTokenManagerInterface $jwtManager,
        EntityManagerInterface $entityManager,
        protected ContainerInterface $container,
        protected TagAwareCacheInterface $cache,
        protected LoggerInterface $logger
    ) {
        $this->userRepository = $entityManager->getRepository(Actor::class);
    }

    /**
     * Called on every request to decide if this authenticator should be
     * used for the request. Returning `false` will cause this authenticator
     * to be skipped.
     */
    public function supports(Request $request): ?bool
    {
        return $request->headers->has('Authorization') && strpos($request->headers->get('Authorization'), 'API-KEY') === 0;
    }

    public function authenticate(Request $request): Passport
    {
        $tokenExtractor = new AuthorizationHeaderTokenExtractor('API-KEY', 'Authorization');

        $apiKeyToken = $tokenExtractor->extract($request);
        return $this->authenticateToken($apiKeyToken, $request);
    }

    public function authenticateToken($apiKeyToken, $request = null): Passport
    {
        if (null === $apiKeyToken) {
            // The token header was empty, authentication fails with HTTP Status
            // Code 401 "Unauthorized"
            throw new CustomUserMessageAuthenticationException(ErrorConstants::FSR_CH_0000014);
        }
        $apiKeyProvider = new ApiKeyProvider($this->container, $this->cache, $this->logger);
        $userId = $apiKeyProvider->validateFingerprint($apiKeyToken, $request);
        $passport = new SelfValidatingPassport(
            new UserBadge($userId, function ($userIdentifier) {
                return $this->userRepository->findOneBy(['userId' => $userIdentifier]);
            })
        );
        $passport->setAttribute('token', $apiKeyToken);

        return $passport;
    }

    /**
     * Shortcut to create a PostAuthenticationToken for you, if you don't really
     * care about which authenticated token you're using.
     */
    public function createToken(Passport $passport, string $firewallName): TokenInterface
    {
        return new HmacPostAuthenticationToken($passport->getUser(), $firewallName, $passport->getUser()->getRoles(true, true), $passport->getAttribute('token'));
    }
}
