<?php
/* Copyright start
  Copyright (C) 2008 - 2024 Fortinet Inc.
  All rights reserved.
  FORTINET CONFIDENTIAL & FORTINET PROPRIETARY SOURCE CODE
  Copyright end */

namespace App\Serializer\Normalizer;

use Psr\Log\LoggerInterface;
use App\Entity\Core\Picklist;
use App\Constants\AppConstants;
use App\Service\MetadataHelper;
use App\Service\UtilityService;
use App\Security\Voter\FsrVoter;
use App\Constants\ErrorConstants;
use App\Constants\ContextConstants;
use Doctrine\ORM\EntityManagerInterface;
use Symfony\Component\PropertyInfo\Type;
use App\Service\NormalizerUtilityService;
use ApiPlatform\Api\IriConverterInterface;
use App\Providers\PermissionsDictionaryBuilder;
use Symfony\Component\HttpFoundation\RequestStack;
use ApiPlatform\Api\ResourceClassResolverInterface;
use Symfony\Contracts\Cache\TagAwareCacheInterface;
use ApiPlatform\Core\Exception\InvalidValueException;
use ApiPlatform\Core\Exception\ItemNotFoundException;
use ApiPlatform\Core\Serializer\AbstractItemNormalizer;
use ApiPlatform\Core\Exception\InvalidArgumentException;
use Symfony\Component\Serializer\Exception\LogicException;
use ApiPlatform\Core\DataProvider\ItemDataProviderInterface;
use Symfony\Component\DependencyInjection\ContainerInterface;
use Symfony\Component\PropertyAccess\PropertyAccessorInterface;
use Symfony\Component\Serializer\Normalizer\NormalizerInterface;
use Symfony\Component\Serializer\Normalizer\DenormalizerInterface;
use Symfony\Component\Serializer\Exception\UnexpectedValueException;
use Symfony\Component\Serializer\NameConverter\NameConverterInterface;
use Symfony\Component\Serializer\Exception\NotNormalizableValueException;
use Symfony\Component\Serializer\Mapping\Factory\ClassMetadataFactoryInterface;
use ApiPlatform\Core\Metadata\Property\Factory\PropertyMetadataFactoryInterface;
use ApiPlatform\Core\Metadata\Resource\Factory\ResourceMetadataFactoryInterface;
use ApiPlatform\Metadata\Property\Factory\PropertyNameCollectionFactoryInterface;

class EfficientItemDeNormalizer extends AbstractItemNormalizer implements NormalizerInterface, DenormalizerInterface
{


    protected $iriConverter;
    protected $classMetadata;
    protected $fsrVoter;
    protected $request;
    protected $normalizerUtilityService;
    protected $containers;
    protected $entityManager;
    protected $metadataHelper;
    protected $requestStack;
    protected $allowPlainIdentifiers;
    private $cache;

    private const REQUIRED_ATTRIBUTE = array("@id", "@context", "@type");
    private const SELF_TENANT = "/api/3/tenants/b3a700f7-00be-4ef9-90c6-3c8fe6e1be63";

    public function __construct(
        PropertyNameCollectionFactoryInterface $propertyNameCollectionFactory,
        PropertyMetadataFactoryInterface $propertyMetadataFactory,
        IriConverterInterface $iriConverter,
        ResourceClassResolverInterface $resourceClassResolver,
        ContainerInterface $containers,
        FsrVoter $fsrVoter,
        EntityManagerInterface $entityManager,
        RequestStack $requestStack,
        NormalizerUtilityService $normalizerUtilityService,
        MetadataHelper $metadataHelper,
        TagAwareCacheInterface $cache,
        protected LoggerInterface $logger,
        protected UtilityService $utilityService,
        PropertyAccessorInterface $propertyAccessor = null,
        NameConverterInterface $nameConverter = null,
        ClassMetadataFactoryInterface $classMetadataFactory = null,
        ItemDataProviderInterface $itemDataProvider = null,
        ResourceMetadataFactoryInterface $resourceMetadataFactory = null,
        array $defaultContext = [],
        iterable $dataTransformers = []
    ) {
        parent::__construct(
            $propertyNameCollectionFactory,
            $propertyMetadataFactory,
            $iriConverter,
            $resourceClassResolver,
            $propertyAccessor,
            $nameConverter,
            $classMetadataFactory,
            null,
            false,
            $defaultContext,
            $dataTransformers,
            $resourceMetadataFactory
        );
        $this->iriConverter = $iriConverter;
        $this->classMetadata = $classMetadataFactory;
        $this->fsrVoter = $fsrVoter;
        $this->requestStack = $requestStack;
        $this->normalizerUtilityService = $normalizerUtilityService;
        $this->containers = $containers;
        $this->entityManager = $entityManager;
        $this->itemDataProvider = $itemDataProvider;
        $this->metadataHelper = $metadataHelper;
        $this->allowPlainIdentifiers = false;
        $this->cache = $cache;
    }

    const FORMAT = 'jsonld';

    /**
     * {@inheritdoc}
     */
    public function supportsDenormalization($data, $type, $format = null, array $context = []): bool
    {
        return self::FORMAT === $format && parent::supportsDenormalization($data, $type, $format);
    }

    public function supportsNormalization($data, $format = null, array $context = []): bool
    {
        return false;
    }

    private function setDefaultValue($context, $data, $class)
    {
        $currentRequest = $this->requestStack->getCurrentRequest();
        if (is_array($data) && array_key_exists('recordTags', $data) && !empty($data['recordTags'])) {
            if (is_string($data['recordTags']) && strlen(trim($data['recordTags'])) <= AppConstants::MIN_TAG_LENGTH) {
                unset($data['recordTags']);
            }
            foreach ($data['recordTags'] as $recordTagKey => $recordTagValue) {
                if (empty(trim($data['recordTags'][$recordTagKey]))) {
                    unset($data['recordTags'][$recordTagKey]);
                }
            }
        }
        if (is_array($data) && array_key_exists('__link', $data) && !empty($data['__link']) && array_key_exists('recordTags', $data['__link'])) {
            if (is_string($data['__link']['recordTags']) && strlen(trim($data['__link']['recordTags']) <= AppConstants::MIN_TAG_LENGTH)) {
                unset($data['__link']['recordTags']);
            } else {
                foreach ($data['__link']['recordTags'] as $recordTagKey => $recordTagValue) {
                    if (empty(trim($data['__link']['recordTags'][$recordTagKey]))) {
                        unset($data['__link']['recordTags'][$recordTagKey]);
                    }
                }
            }
        }
        array_values($data);
        if ((isset($context['_is_update']) && $context['_is_update'] === true) || ($currentRequest && $currentRequest->getMethod() == 'PUT')) {
            // Default Value should not be set for Update Case.
            // It's Creating issues at places where we are not sending the whole objects in case of update.
            // One Common Use case is postman updating the agent status
            return $data;
        }

        $entityFieldMap = $this->metadataHelper->getEntityFields($class) ?? null;
        $defaultValues = $entityFieldMap['defaultValue'];
        foreach ($defaultValues as $fieldName => $defaultValue) {
            if (array_key_exists($fieldName, $data)) {
                if (is_null($data[$fieldName]) || $data[$fieldName] === '') {
                    if (
                        array_key_exists($fieldName, $entityFieldMap['attributeFormType'])
                        && $entityFieldMap['attributeFormType'][$fieldName] == 'datetime'
                        && is_string($defaultValue)
                        && strpos($defaultValue, AppConstants::CUSTOM_DATE_DEFAULT_VALUE_IDENTIFIER)
                    ) {
                        continue;
                    }
                    $data[$fieldName] = $defaultValue;
                }
            } else {
                if (
                    array_key_exists($fieldName, $entityFieldMap['attributeFormType'])
                    && $entityFieldMap['attributeFormType'][$fieldName] == 'datetime'
                    && is_string($defaultValue)
                    && strpos($defaultValue, AppConstants::CUSTOM_DATE_DEFAULT_VALUE_IDENTIFIER)
                ) {
                    continue;
                }
                $data[$fieldName] = $defaultValue;
            }
        }
        return $data;
    }

    /**
     * {@inheritdoc}
     *
     * @throws NotNormalizableValueException
     */
    public function denormalize($data, $class, $format = null, array $context = [])
    {
        $context[ContextConstants::FSR_MODULE_CLASS_NAME] = $class;

        if (strpos($class, 'App\\Entity\\Custom\\') === 0) {
            $data = $this->setDefaultValue($context, $data, $class);
        }
        // Avoid issues with proxies if we populated the object
        if (isset($data['@id']) && !isset($context[self::OBJECT_TO_POPULATE])) {
            if (true !== ($context['api_allow_update'] ?? true)) {
                throw new NotNormalizableValueException(ErrorConstants::FSR_CH_0000005);
            }

            $context[self::OBJECT_TO_POPULATE] = $this->iriConverter->getResourceFromIri($data['@id'], $context + ['fetch_data' => true]);
        }
        // type validation is disabled by default unless specifically enabled in context
        if (!isset($context[self::DISABLE_TYPE_ENFORCEMENT])) {
            $context[self::DISABLE_TYPE_ENFORCEMENT] = true;
        }
        foreach ($data as $indexKey => $objectValue) {
            if (isset($data[$indexKey]) && is_array($data[$indexKey]) && array_key_exists('@type', $data[$indexKey]) && $data[$indexKey]['@type'] == 'File') {
                $data[$indexKey] = $data[$indexKey]['@id'];
            }
        }
        return parent::denormalize($data, $class, $format, $context);
    }

    private function supportsPlainIdentifiers(): bool
    {
        return $this->allowPlainIdentifiers && null !== $this->itemDataProvider;
    }

    /**
     * Denormalizes a relation.
     *
     * @throws LogicException
     * @throws UnexpectedValueException
     * @throws ItemNotFoundException
     *
     * @return object|null
     */
    protected function denormalizeRelation(string $attributeName, $propertyMetadata, string $className, $value, ?string $format, array $context)
    {
        $supportsPlainIdentifiers = $this->supportsPlainIdentifiers();
        $isPicklist = ($className === Picklist::class || $className == AppConstants::PICKLIST_CLASS);

        if (!$propertyMetadata->getType()->isCollection() && is_array($value) && array_key_exists('@id', $value) && $value["@id"]) {
            if (!$isPicklist) return $this->iriConverter->getResourceFromIri($value["@id"], $context + ['fetch_data' => false]);
            else $value = $value["@id"];
        }

        $iri = $this->iriConverter->getIriFromResource($className);

        // validate picklist value
        if ($isPicklist && is_array($value)  && array_key_exists(AppConstants::PICKLIST_VALUE, $value)) {
            if (!array_key_exists(AppConstants::PICKLIST_NAME, $value)) {
                $value[AppConstants::PICKLIST_NAME] = $this->utilityService->getPicklistNameByAttribute($attributeName, $context);
            }
            return $this->metadataHelper->handlePicklistNameToIriConversion($iri, $value, $context);
        }

        if ($isPicklist && !is_array($value) && !is_null($value)) {
            $value = trim($value);
            if (strlen($value) > 0) $this->utilityService->validatePicklistValue($context, $attributeName, $iri, $value);
        }

        if (\is_string($value)) {
            try {
                return $this->iriConverter->getResourceFromIri($value, $context + ['fetch_data' => false]);
            } catch (InvalidArgumentException $e) {
                try {
                    $valueIRI = sprintf('%s/%s', $iri, $value);
                    return $this->iriConverter->getResourceFromIri($valueIRI, $context + ['fetch_data' => false]);
                } catch (InvalidArgumentException $e) {
                    return null;
                    // ignoring an invalid iri for backward compatibility
                    // throw new HttpException(Response::HTTP_BAD_REQUEST, (sprintf(
                    //     'IRI not supported (found "%s" in "%s" of "%s")',
                    //     $value,
                    //     $attributeName,
                    //     $className
                    // )), $e);
                } catch (\Exception $e) {
                    $this->logger->error(sprintf(ErrorConstants::FSR_CH_0000006, $value));
                }
                if (!$supportsPlainIdentifiers) {
                    throw new UnexpectedValueException(sprintf(ErrorConstants::FSR_CH_0000006, $value), $e->getCode(), $e);
                }
            } catch (\Exception $e) {
                return null;
            }
        }

        if ($propertyMetadata->isWritableLink()) {
            $context['api_allow_update'] = true;

            if (!$this->serializer instanceof DenormalizerInterface) {
                throw new LogicException(sprintf(ErrorConstants::FSR_CH_0000007, DenormalizerInterface::class));
            }

            try {
                $item = $this->serializer->denormalize($value, $className, $format, $context);
                if (!\is_object($item) && null !== $item) {
                    throw new UnexpectedValueException(ErrorConstants::FSR_CH_0000008);
                }

                return $item;
            } catch (InvalidValueException $e) {
                if (!$supportsPlainIdentifiers) {
                    throw $e;
                }
            } catch (\Exception $e) {
                if (!$supportsPlainIdentifiers) {
                    throw $e;
                }
            }
        }

        if (!\is_array($value)) {
            if (!$supportsPlainIdentifiers) {
                throw new UnexpectedValueException(sprintf(ErrorConstants::FSR_CH_0000009, $attributeName, \gettype($value)));
            }

            $item = $this->itemDataProvider->getItem($className, $value, null, $context + ['fetch_data' => true]);
            if (null === $item) {
                throw new ItemNotFoundException(sprintf(ErrorConstants::FSR_CH_0000010, $className, $value));
            }

            return $item;
        }

        throw new UnexpectedValueException(sprintf(ErrorConstants::FSR_CH_0000011, $attributeName));
    }

    protected function getAllowedAttributes($classOrObject, array $context, $attributesAsString = false)
    {
        // get_class($classOrObject) returns false in case if its an class not an object
        // Adding ternary here elseif returns false
        // purpose is to check if object retrieve its class else directly consume the class
        $entity = is_string($classOrObject) ? $classOrObject : get_class($classOrObject);
        $module = $this->metadataHelper->getModuleFromClassName($entity);
        $allowedAttributes = [];
        $currentRequest = $this->requestStack->getCurrentRequest();

        $is_update = false;
        if ((isset($context['_is_update']) && $context['_is_update'] == true) || ($currentRequest && $currentRequest->getMethod() == 'PUT')) {
            $is_update = true;
        }
        $context['ignore_authorization'] ?? $context['ignore_authorization'] = false;
        $context['ignore_field_authorization'] ?? $context['ignore_field_authorization'] = false;
        $context['ignore_relationship_authorization'] ?? $context['ignore_relationship_authorization'] = false;
        $options = $this->getFactoryOptions($context);
        $propertyNames = $this->propertyNameCollectionFactory->create($context['resource_class'], $options);
        $prev_data = [];
        if ($currentRequest) {
            $prev_data = $currentRequest->attributes->get('fsr_previous_data');
        }
        foreach ($propertyNames as $attributeName => $attributeMetadata) {
            $propertyMetadata = $this->propertyMetadataFactory->create($classOrObject, $attributeMetadata);
            $attribute_prev_value = $prev_data[$attributeMetadata] ?? null;
            $propertyAdded = false;
            if (
                $propertyMetadata->isWritable() || !$is_update
                || ($attribute_prev_value == null || ($attributeMetadata == "tenant" && $attribute_prev_value == self::SELF_TENANT))
            ) {
                if (
                    $context['ignore_authorization']
                    || $context['ignore_field_authorization']
                    || $context['ignore_relationship_authorization']
                    || $this->fsrVoter->isGranted(PermissionsDictionaryBuilder::PERM_FIELD_WRITE . '.' . $module . '.' . $attributeMetadata)
                ) {
                    $allowedAttributes[] = $attributeMetadata;
                    $propertyAdded = true;
                }
            }

            // assign data to allow mandatory attributes also even it is not accessible to use
            if (!$propertyAdded && !$is_update && $propertyMetadata->isRequired() && $attribute_prev_value == null) {
                $allowedAttributes[] = $attributeMetadata;
            }
        }
        $allowedAttributes = array_merge($allowedAttributes, self::REQUIRED_ATTRIBUTE);
        return $allowedAttributes;
    }

    /**
     * Denormalizes a collection of objects.
     *
     * @throws InvalidArgumentException
     */
    protected function denormalizeCollection(string $attribute, $propertyMetadata, Type $type, string $className, $value, ?string $format, array $context): array
    {
        if (is_null($value)) {
            return [];
        }

        if (is_string($value) && strlen(trim($value)) === 0) {
            return [];
        }

        if (!\is_array($value)) {
            $value = explode(",", $value);
        }

        /* Validate if request is coming for custom modules
           Also validate if ignore missing relationship params is set over requests.
           If set, we will validate the related record against database and check there validity.
           This will be done for all ManyToMany Fields.
        */
        $currentRequest = $this->requestStack->getCurrentRequest();
        $queryParams = $currentRequest ? $currentRequest->query->all() : [];
        if (
            $currentRequest && array_key_exists(AppConstants::IGNORE_MISSING_RELATIONSHIP, $queryParams) &&
            $queryParams[AppConstants::IGNORE_MISSING_RELATIONSHIP] &&
            str_starts_with($className, AppConstants::CUSTOM_MODULE_IDENTIFIER)
        ) {
            $value = $this->utilityService->validateRelationFieldsValues($className, $value);
        }
        return parent::denormalizeCollection($attribute, $propertyMetadata, $type, $className, $value, $format, $context);
    }
}
