<?php
/* Copyright start
  Copyright (C) 2008 - 2024 Fortinet Inc.
  All rights reserved.
  FORTINET CONFIDENTIAL & FORTINET PROPRIETARY SOURCE CODE
  Copyright end */

namespace App\Service\ConfigExportImport;

use App\Constants\MessageConstants;
use App\Entity\Core\ExportTemplate;
use App\Providers\PermissionsDictionaryBuilder;
use App\Security\Voter\FsrVoter;
use App\Service\BulkRequestService;
use App\Service\UtilityService;
use DateTime;
use Doctrine\ORM\EntityManagerInterface;
use Psr\Log\LoggerInterface;
use stdClass;
use Symfony\Component\DependencyInjection\ContainerInterface;
use Symfony\Component\HttpFoundation\JsonResponse;
use Symfony\Component\HttpFoundation\Response;
use Symfony\Component\Security\Core\Authentication\Token\Storage\TokenStorageInterface;
use Symfony\Component\Serializer\Normalizer\NormalizerInterface;

class ExportTemplateConfig extends PortingConfig
{
    const EXPORT_TEMPLATE_URL = "/api/3/export_templates";

    public function __construct(
        private BulkRequestService $bulkRequestService,
        private FsrVoter $fsrVoter,
        EntityManagerInterface $entityManager,
        NormalizerInterface $normalizer,
        TokenStorageInterface $tokenStorage,
        ContainerInterface $container,
        private UtilityService $utilityService,
        protected LoggerInterface $logger
    ) {
        parent::__construct($entityManager, $normalizer, $tokenStorage, $container, $logger);
    }

    public function generateInfoContent(array $options)
    {
        $entityRepo = $this->entityManager->getRepository(ExportTemplate::class);
        return $this->utilityService->getItemWithWhere($entityRepo, ['uuid' => $options], selectFields: ['o.name']);
    }

    public function getImportOptions(array $data, string $key, $tokenType): array
    {
        // check permission
        if (!$this->checkPermission($this->fsrVoter, 'security')) return [];
        $options = [];
        $exportTemplateRepository = $this->entityManager->getRepository(ExportTemplate::class);
        $existingExportTemplates = $exportTemplateRepository->findAll();
        $context = $this->getDefaultNormalizationContext(ExportTemplate::class, false);
        $normalizedData = $this->normalizer->normalize($existingExportTemplates, 'jsonld', $context);
        $normalizedData = array_key_exists('hydra:member', $normalizedData) && $normalizedData['hydra:member'] ? $normalizedData['hydra:member'] : [];
        foreach ($data as $exportTemplate) {
            $templateOptions = [];
            $existingExportTemplate = self::findBy($normalizedData, 'name', $exportTemplate['name'], true);
            $templateOptions['exists'] = $existingExportTemplate ? true : false;
            $templateOptions['include'] = true;
            $templateOptions['name'] = $exportTemplate['name'];
            if ($existingExportTemplate) {
                $templateOptions['@id'] = $existingExportTemplate['@id'];
            }
            $options[] = $templateOptions;
        }
        return $options;
    }

    /**
     * Generates import options from the tmp folder
     *
     * @param string $key
     * @param string $tmpDirectory
     * @param string $tokenType
     * @return array
     */
    public function getImportOptionsFromFile(string $key, string $tmpDirectory, $tokenType): array
    {
        $data = $this->loadDataFromDirectory($key, $tmpDirectory);
        return $this->getImportOptions($data, $key, $tokenType);
    }

    protected function loadDataFromDirectory(string $key, string $tmpDirectory): array
    {
        $exportTemplatefolderPath = $tmpDirectory . $key . '/';
        $exportTemplates = [];
        if (file_exists($exportTemplatefolderPath)) {
            $exportTemplates = scandir($exportTemplatefolderPath);
        }
        $data = [];
        foreach ($exportTemplates as $exportTemplate) {
            // Continuing loop by ignoring current and parrent directory coming when dir is scanned for export templates
            if (in_array($exportTemplate, ['.', '..'])) {
                continue;
            }
            $exportTemplatePath = $exportTemplatefolderPath . $exportTemplate;

            if (file_exists($exportTemplatePath)) {
                $fileData = file_get_contents($exportTemplatePath);
                $jsonData = json_decode($fileData, true);
                $data[] = $jsonData;
            }
        }
        return $data;
    }

    public function import(array $data, array $options, $tokenType, array $importedBy): bool
    {
        $optionMap = $this->prepareOptionMap($options);

        list($exportTemplatesToAdd, $exportTemplatesToUpdate) = $this->filterRecordsByOperation($data, $optionMap, $importedBy);

        $createResult = $this->createExportTemplateRequest($exportTemplatesToAdd, $tokenType);
        $updateResult = $this->updateExportTemplateRequest($exportTemplatesToUpdate, $tokenType);

        return $createResult && $updateResult;
    }

    /**
     * @param array $options
     * @return array
     */
    protected function prepareOptionMap(array $options): array
    {
        //Prepare option map
        $optionMap = array();
        foreach ($options as $option) {
            // Skip importing exportTemplates which are not choose to import
            if (array_key_exists('include', $option) && !$option['include']) {
                continue;
            }
            $optionMap[$option['name']] = $option;
        }
        return $optionMap;
    }

    /**
     * @param array $data
     * @param array $optionMap
     * @param mixed $importedBy
     * @return array[]
     */
    public function filterRecordsByOperation(array $data, array $optionMap, mixed $importedBy): array
    {
        $exportTemplatesToAdd = [];
        $exportTemplatesToUpdate = [];

        $currentUser = $this->getUser();
        foreach ($data as $exportTemplate) {
            $timestamp = (new \DateTime('NOW', new \DateTimeZone('UTC')))->getTimestamp();
            if (!array_key_exists($exportTemplate['name'], $optionMap))
                continue;

            $exportTemplateOption = $optionMap[$exportTemplate['name']];
            $exportTemplateExists = $exportTemplateOption['exists'];

            $exportTemplate['modifyUser'] = $currentUser->getUuid();
            $exportTemplate['modifyDate'] = $timestamp;

            if (array_key_exists('uuid', $exportTemplate)) {
                unset($exportTemplate['uuid']);
            }

            // Export templates exists with name
            if ($exportTemplateExists) {
                if (array_key_exists('@id', $exportTemplateOption)) {
                    $exportTemplate['@id'] = $exportTemplateOption['@id'];
                }
                $exportTemplatesToUpdate[] = (object) $exportTemplate;
            } else {
                $exportTemplate['createUser'] = $currentUser->getUuid();
                $exportTemplate['createDate'] = $timestamp;
                if (array_key_exists('@id', $exportTemplate)) {
                    unset($exportTemplate['@id']);
                }
                $exportTemplatesToAdd[] = $exportTemplate;
            }
        }
        return array($exportTemplatesToAdd, $exportTemplatesToUpdate);
    }

    /**
     * @param array $exportTemplatesToAdd
     * @param $tokenType
     * @return bool
     */
    public function createExportTemplateRequest(array $exportTemplatesToAdd, $tokenType): bool
    {
        $success = true;
        if (sizeof($exportTemplatesToAdd) > 0) {
            $addObj = new stdClass();
            $addObj->{'data'} = $exportTemplatesToAdd;
            $this->addMessage(MessageConstants::PREPARING_CREATE_REQUEST);
            $createResults = $this->triggerBulkRequest("POST", $addObj, $tokenType);
            $this->addMessage(MessageConstants::NEW_RECORDS_CREATED);
            if ($createResults->getStatusCode() != Response::HTTP_CREATED)
                $success = false;
        }
        return $success;
    }

    /**
     * @param array $exportTemplatesToUpdate
     * @param $tokenType
     * @return bool
     */
    public function updateExportTemplateRequest(array $exportTemplatesToUpdate, $tokenType): bool
    {
        $success = true;
        if (sizeof($exportTemplatesToUpdate) > 0) {
            $updateObject = new stdClass();
            $updateObject->{'data'} = $exportTemplatesToUpdate;
            $this->addMessage(MessageConstants::PREPARING_UPDATE_REQUEST);
            $updateResults = $this->triggerBulkRequest("PUT", $updateObject, $tokenType);
            $this->addMessage(MessageConstants::RECORDS_UPDATED);
            if ($updateResults->getStatusCode() != Response::HTTP_OK)
                $success = false;
        }
        return $success;
    }

    protected function triggerBulkRequest($method, $data, $tokenType): JsonResponse
    {
        $headers = [];
        $token = $this->getTokenFromStorage();
        $tokenPrefix = $tokenType == 'JWT' ? 'Bearer ' : 'CS ';
        $headers['Authorization'] = $token ? $tokenPrefix . $token->getCredentials() : null;
        $headers['Content-Type'] = 'application/json';

        return $this->bulkRequestService->prepareSubRequest(
            ExportTemplate::class,
            $data,
            self::EXPORT_TEMPLATE_URL,
            $headers,
            [],
            $method,
            true,
            'import'
        );
    }

    /**
     * Imports from the tmp folder
     *
     * @param array $options
     * @param string $key
     * @param string $tmpDirectory
     * @param [type] $tokenType
     * @return void
     */
    public function importFromFile(array $options, string $key, string $tmpDirectory, $tokenType, $importedBy = []): bool
    {
        $data = $this->loadDataFromDirectory($key, $tmpDirectory);
        return $this->import($data, $options, $tokenType, $importedBy);
    }

    public function generateExportOptions(array $options, $tokenType = 'JWT')
    {
        $exportOption = [];
        if (is_iterable($options)) {
            foreach ($options as $value) {
                $value['value'] = $value['uuid'];
                array_push($exportOption, $value);
            }
        }
        return $exportOption;
    }

    public function export(array $options, $tokenType = 'JWT')
    {
        if (!$this->fsrVoter->isGranted(PermissionsDictionaryBuilder::PERM_CRUD_READ . '.security')) {
            return [];
        }
        return $this->getResourceFromIRI($options, ExportTemplate::class);
    }

    /**
     * Exports the config using the specified options
     *
     * @param array $options
     * @return array
     */
    public function exportFile(array $options, $type, $tmpDirectory, $tokenType = 'JWT')
    {
        $modifiedOptions = [];

        foreach ($options as $value) {
            if (is_string($value)) {
                $modifiedOptions[] = $value;
            }
            else if (is_array($value) && !empty($value)) {
                $modifiedOptions[] = $value['value'];
            }
        }
        $data = $this->export($modifiedOptions, $tokenType);
        if (!$data) {
            return $data;
        }
        foreach ($data as $item) {
            $templateName = $this->cleanFileName($item['name']);
            $filename = $tmpDirectory . $type . '/' . $templateName . '.json';
            $this->writeToFile($filename, $item);
            $filenames[] = $templateName;
        }
        $this->appendToFile($tmpDirectory . 'info.json', $type, $filenames);
        return $filenames;
    }
}
