<?php
/* Copyright start
  Copyright (C) 2008 - 2023 Fortinet Inc.
  All rights reserved.
  FORTINET CONFIDENTIAL & FORTINET PROPRIETARY SOURCE CODE
  Copyright end */

namespace App\Service\ConfigExportImport;

use App\Constants\MessageConstants;
use App\Entity\Rules\PreprocessingRule;
use App\Providers\PermissionsDictionaryBuilder;
use App\Security\Voter\FsrVoter;
use App\Service\BulkRequestService;
use Doctrine\ORM\EntityManagerInterface;
use Psr\Log\LoggerInterface;
use stdClass;
use Symfony\Component\DependencyInjection\ContainerInterface;
use Symfony\Component\HttpFoundation\Response;
use Symfony\Component\Security\Core\Authentication\Token\Storage\TokenStorageInterface;
use Symfony\Component\Serializer\Normalizer\NormalizerInterface;

/**
 * Class RuleConfig
 * @package App\Service\ConfigExportImport
 */
class PreprocessingRulesConfig extends PortingConfig
{
    const MODULE_NAME = "preprocessing_rules";
    const RULE_URL = "/api/3/preprocessing_rules";


    public function __construct(
        private BulkRequestService $bulkRequestService,
        private FsrVoter           $fsrVoter,
        EntityManagerInterface     $entityManager,
        NormalizerInterface        $normalizer,
        TokenStorageInterface      $tokenStorage,
        ContainerInterface         $container,
        protected LoggerInterface  $logger
    )
    {
        parent::__construct($entityManager, $normalizer, $tokenStorage, $container, $logger);
    }

    private function getUUIDsFromOptions(array $options)
    {
        return !empty($options) && is_string($options[0]) ? $options :
            array_map(function ($item) {
                return $item['value'];
            }, $options);
    }

    /**
     * Exports the config using the specified options
     *
     * @param array $options
     * @return array
     */
    public function export(array $options, $tokenType = 'JWT')
    {
        if (!$this->fsrVoter->isGranted(PermissionsDictionaryBuilder::PERM_CRUD_READ . '.preprocessing_rules')) {
            return [];
        }
        if (!empty($options)) {
            $uuids = $this->getUUIDsFromOptions($options);
            $ignoredAttributes = ['id', 'createUser', 'modifyUser', 'sattrib', 'importedBy', 'lastModifyDate'];
            $preprocessingRules = $this->getResourceFromIRI($uuids, PreprocessingRule::class, ignoredAttributes: $ignoredAttributes);
        }
        return $preprocessingRules;

    }

    /**
     * Exports the config to a tmp file using the specified options
     *
     * @param array $options
     * @param string $type
     * @param string $tokenType
     */
    public function exportFile(array $options, $type, $tmpDirectory, $tokenType = 'JWT')
    {
        $metaData = [];
        $data = $this->export($options, $tokenType);
        if (!$data) {
            return $data;
        }
        $directory = $tmpDirectory . $type . '/';
        foreach ($data as $item) {
            $name = $this->cleanFileName($item['name']);
            $filename = $directory . $name . '.json';
            $this->writeToFile($filename, $item);
            $filenames[] = $filename;
            $metaData[] = $item['name'];
        }
        $this->appendToFile($tmpDirectory . 'info.json', $type, $metaData);
        return $filenames;
    }

    /**
     * Generates import options from the tmp folder
     *
     * @param string $key
     * @param string $tmpDirectory
     * @param string $tokenType
     * @return array
     */
    public function getImportOptionsFromFile(string $key, string $tmpDirectory, $tokenType)
    {
        $data = $this->loadDataFromDirectory($key, $tmpDirectory);
        return $this->getImportOptions([$key => $data], $key, $tokenType);
    }

    /**
     * @param array $allData
     * @param string $key
     * @return array
     */
    public function getImportOptions(array $allData, string $key, $tokenType)
    {
        $data = $allData[$key];
        $result = [];

        // check permission
        if (!$this->checkPermission($this->fsrVoter, self::MODULE_NAME)) return $result;

        if ($key == 'preprocessingRules' && !empty($data)) {
            $repository = $this->entityManager->getRepository(PreprocessingRule::class);
            $options = [];
            foreach ($data as $rule) {
                $existingRecord = $repository->findOneBy(['uuid' => $rule['uuid']]);
                $option = [];
                $option['name'] = $rule['name'];
                $option['uuid'] = $existingRecord ? $existingRecord->getUuid() : $rule['uuid'];
                $option['include'] = ($this->importJob && $this->importJob->getType() === 'SolutionPack Import') || !$existingRecord;
                $option['exists'] = (bool)$existingRecord;
                $options[] = $option;
            }
            $result = $options;
        }
        return $result;
    }

    public function generateExportOptions(array $options, $tokenType = 'JWT')
    {
        $exportOption = [];
        if (is_iterable($options)) {
            foreach ($options as $value) {
                $value['value'] = $value['uuid'];
                array_push($exportOption, $value);
            }
        }
        return $exportOption;
    }

    public function generateInfoContent(array $options)
    {
        $content = [];
        foreach ($options as $preprocessingRule) {
            $content[] = $preprocessingRule["name"];
        }

        return $content;
    }

    /**
     * Imports the config data using the specified options
     *
     * @param array $data
     * @param array $options
     * @param $tokenType
     * @param array $importedBy
     * @return bool
     */
    public function import(array $data, array $options, $tokenType, array $importedBy = []): bool
    {
        $optionMap = $this->prepareOptionMap($options);

        list($rulesToAdd, $rulesToUpdate) = $this->filterRecordsByOperation($data, $optionMap, $importedBy);

        $createResult = $this->createRuleRequest($rulesToAdd, $tokenType);
        $updateResult = $this->updateRuleRequest($rulesToUpdate, $tokenType);

        return $createResult && $updateResult;
    }

    public function triggerBulkRequest($method, $data, $tokenType): \Symfony\Component\HttpFoundation\JsonResponse
    {
        $headers = [];
        $token = $this->getTokenFromStorage();
        $tokenPrefix = $tokenType == 'JWT' ? 'Bearer ' : 'CS ';
        $headers['Authorization'] = $token ? $tokenPrefix . $token->getCredentials() : null;
        $headers['Content-Type'] = 'application/json';

        return $this->bulkRequestService->prepareSubRequest(
            PreprocessingRule::class,
            $data,
            PreprocessingRulesConfig::RULE_URL,
            $headers,
            [],
            $method,
            true,
            'import'
        );
    }

    /**
     * Imports from the tmp folder
     *
     * @param array $options
     * @param string $key
     * @param string $tmpDirectory
     * @param [type] $tokenType
     * @param array $importedBy
     * @return void
     */
    public function importFromFile(array $options, string $key, string $tmpDirectory, $tokenType, $importedBy = []): bool
    {
        $data = $this->loadDataFromDirectory($key, $tmpDirectory);
        return $this->import($data, $options, $tokenType, $importedBy);
    }

    /**
     * @param array $data
     * @param array $optionMap
     * @param mixed $importedBy
     * @return array[]
     */
    public function filterRecordsByOperation(array $data, array $optionMap, mixed $importedBy): array
    {
        $rulesToAdd = [];
        $rulesToUpdate = [];

        $currentUser = $this->getUser();
        foreach ($data as $rule) {
            if (!array_key_exists($rule['uuid'], $optionMap)) continue;

            $option = $optionMap[$rule['uuid']];
            $ruleAlreadyExists = $option['exists'];

            $rule['importedBy'] = !empty($rule['importedBy']) ? array_merge($rule['importedBy'], $importedBy) : $importedBy;
            $rule['modifyUser'] = $currentUser->getUuid();


            if ($ruleAlreadyExists) $rulesToUpdate[] = (object)$rule;
            else {
                $rule['createUser'] = $currentUser->getUuid();
                $rulesToAdd[] = $rule;
            }
        }
        return array($rulesToAdd, $rulesToUpdate);
    }

    /**
     * @param array $rulesToAdd
     * @param $tokenType
     * @return bool
     */
    public function createRuleRequest(array $rulesToAdd, $tokenType): bool
    {
        $success = true;
        if (sizeof($rulesToAdd) > 0) {
            $addObj = new stdClass();
            $addObj->{'data'} = $rulesToAdd;
            $this->addMessage(MessageConstants::PREPARING_CREATE_REQUEST);
            $createResults = $this->triggerBulkRequest("POST", $addObj, $tokenType);
            $this->addMessage(MessageConstants::NEW_RECORDS_CREATED);
            if ($createResults->getStatusCode() != Response::HTTP_CREATED) $success = false;
        }
        return $success;
    }

    /**
     * @param array $rulesToUpdate
     * @param $tokenType
     * @return bool
     */
    public function updateRuleRequest(array $rulesToUpdate, $tokenType): bool
    {
        $success = true;
        if (sizeof($rulesToUpdate) > 0) {
            $updateObject = new stdClass();
            $updateObject->{'data'} = $rulesToUpdate;
            $this->addMessage(MessageConstants::PREPARING_UPDATE_REQUEST);
            $updateResults = $this->triggerBulkRequest("PUT", $updateObject, $tokenType);
            $this->addMessage(MessageConstants::RECORDS_UPDATED);
            if ($updateResults->getStatusCode() != Response::HTTP_OK) $success = false;
        }
        return $success;
    }

    /**
     * @param array $options
     * @return array
     */
    public function prepareOptionMap(array $options): array
    {
//prepare option map
        $optionMap = array();
        foreach ($options as $option) {
            if (!$option['include']) continue;
            $optionMap[$option['uuid']] = $option;
        }
        return $optionMap;
    }

}