<?php
/* Copyright start
  Copyright (C) 2008 - 2024 Fortinet Inc.
  All rights reserved.
  FORTINET CONFIDENTIAL & FORTINET PROPRIETARY SOURCE CODE
  Copyright end */

namespace App\Service;

use App\Constants\AppConstants;
use App\Service\MetadataHelper;
use App\Constants\AttributeLevelOperators;
use Symfony\Component\DependencyInjection\ParameterBag\ParameterBagInterface;
use Symfony\Component\HttpFoundation\Request;
use Symfony\Component\HttpFoundation\RequestStack;

class NormalizerUtilityService
{
    protected $requestStack;

    protected $allowedAttributes = [];

    private $metadataHelper;

    private $parameters;

    const POSTMAN_SECRET_KEY = 'eW1hSe7mnlIi1a0mmgno1nSe9Rea7pphG2oAdmCeBAoduYnd';

    public function __construct(RequestStack $requestStack, MetadataHelper $metadataHelper, ParameterBagInterface $parameters)
    {
        $this->requestStack = $requestStack;
        $this->metadataHelper = $metadataHelper;
        $this->parameters = $parameters;
    }

    public function validateRequestKey($context)
    {
        $request = $this->requestStack->getCurrentRequest();
        if ($request) {
            $postman_secret_key = self::POSTMAN_SECRET_KEY;
            if (!$request->query->has("\$secretKey")) {
                return false;
            } else {
                $SECRET_KEY = $request->query->get("\$secretKey");
                if ($SECRET_KEY === $postman_secret_key) {
                    return true;
                }
            }
        }
        return false;
    }

    public function getResponseIdentifier(&$context)
    {
        $relationships = $this->getOption(AppConstants::NORMALIZER_RELATIONSHIPS, $context);
        $export = $this->getOption(AppConstants::NORMALIZER_EXPORT, $context);
        $asJson = $this->getOption(AppConstants::NORMALIZER_AS_JSON, $context);
        $archived = $this->getOption(AppConstants::NORMALIZER_ARCHIVED, $context);
        $fsrMaxRelationCount = $this->getOption(AppConstants::NORMALIZER_FSR_MAX_RELATIONSHIP_COUNT, $context, $this->getDefaultRelCount(), false);
        $showPassword = (array_key_exists(AppConstants::NORMALIZER_SHOW_PASSWORD, $context) && $context['showPassword'] === true) ? $context['showPassword'] : $this->validateRequestKey($context);
        if ($showPassword === true) {
            $context['showPassword'] = true;
        }
        $hideUuid = $this->getOption(AppConstants::NORMALIZER_HIDE_UUID, $context) ?? false;
        $displayNameOnly = $this->getOption(AppConstants::NORMALIZER_DISPLAY_NAME_ONLY, $context) ?? false;
        $csvNeutralization = $this->getOption(AppConstants::NORMALIZER_CSV_NEUTRALIZATION, $context) ?? false;
        $humanReadableDateFormat = $this->getOption(AppConstants::NORMALIZER_HUMAN_READABLE_DATE_FORMAT, $context) ?? false;
        return [
            $relationships, $export, $asJson,
            $showPassword, $hideUuid, $archived,
            $fsrMaxRelationCount, $displayNameOnly, $csvNeutralization,
            $humanReadableDateFormat
        ];
    }

    private function getSelectedFields()
    {
        $request = $this->requestStack->getCurrentRequest();
        if ($request) {
            if ($request->getMethod() == Request::METHOD_POST) {
                $requestBody = json_decode($request->getContent(), true);
                if (isset($requestBody['__selectFields']) && is_array($requestBody['__selectFields'])) {
                    // Adding UUID as for normalization of relation required minimal of one context to match
                    return array_merge($requestBody['__selectFields'], ["uuid", "id"]);
                } elseif (array_key_exists("__selectFields", $request->query->all())) {
                    return array_merge(explode(',', $request->query->get("__selectFields")), ["uuid", "id"]);
                }
            } elseif ($request->getMethod() == Request::METHOD_GET && array_key_exists("__selectFields__", $GLOBALS)) {
                return array_merge(explode(',', $GLOBALS['__selectFields__']), ["uuid", "id"]);
            }
        }

        return null;
    }
    private function getIgnoreFields()
    {
        $request = $this->requestStack->getCurrentRequest();
        if ($request) {
            if ($request->getMethod() == Request::METHOD_POST) {
                $requestBody = json_decode($request->getContent(), true);
                if (isset($requestBody['__ignoreFields']) && is_array($requestBody['__ignoreFields']) && !empty($requestBody['__ignoreFields'])) {
                    return $requestBody['__ignoreFields'];
                }
            } elseif ($request->getMethod() == Request::METHOD_GET && array_key_exists("__ignoreFields__", $GLOBALS)) {
                return $GLOBALS['__ignoreFields__'];
            }
        }

        return null;
    }

    public function manipulateContextGroup(
        $context,
        $relationships = false,
        $export = false,
        $asJson = false,
        $showPassword = false,
        $archived = false,
        $fsr_max_relation_count = 0,
        $displayNameOnly = false,
        $csvNeutralization = false,
        $humanReadableDateFormat=false
    ) {
        if ($relationships) {
            $context["groups"] = array_key_exists("groups", $context) ? ["fsr_primary"] : ["fsr_all"];
        } else {
            $context["groups"] = ["fsr_primary"];
        }
        if ($showPassword) {
            $context["groups"] = array_merge($context["groups"], ["password"]);
        }
        if ($this->getSelectedFields() && count($context["object_class"]) == 1) {
            $context["groups"] = $this->getSelectedFields();
        }
        if ($this->getIgnoreFields() && count($context["object_class"]) == 1) {
            $context["ignored_attributes"] = $this->getIgnoreFields();
        }
        if (count($context["object_class"]) > 1 && ($archived || (array_key_exists('displayNameOnly', $context) && $context['displayNameOnly']))) {
            $context["groups"] = $this->metadataHelper->getDisplayFieldsForModule($context);
        }
        if ($displayNameOnly) {
            $context["displayNameOnly"] = true;
        }
        if ($csvNeutralization) {
            $context["csvNeutralization"] = true;
        }
        if ($fsr_max_relation_count) {
            $context['fsr_max_relation_count'] = (int)$fsr_max_relation_count;
        }
        if ($humanReadableDateFormat) {
            $context['humanReadableDateFormat'] = $humanReadableDateFormat;
        }

        return $context;
    }

    /**
     * @param $optionName
     * @param array $context
     * @param bool $default
     * @param bool $validateBoolean
     * @return bool|mixed
     */
    public function getOption($optionName, array $context, $default = false, $validateBoolean = true)
    {
        $option = $default;
        $request = $this->requestStack->getCurrentRequest();
        if (!is_null($request)) {
            $key = AttributeLevelOperators::OPERATOR_TOKEN . $optionName;
            $rawOption = $request->query->get($key);
            if (is_null($rawOption)) {
                $rawOption = $request->request->get($key, $default);
            }
            $option = $rawOption;
            if ($validateBoolean) {
                $option = filter_var($rawOption, FILTER_VALIDATE_BOOLEAN);
            }
        }

        $option = array_key_exists($optionName, $context)
            ? $context[$optionName]
            : $option;

        return $option;
    }


    /**
     * @return int
     */
    private function getDefaultRelCount():int
    {
        $val = 1;

        if($this->parameters->has(AppConstants::CONFIG_MAX_RESPONSE_REL_COUNT)){
            $configVal = (int) $this->parameters->get(AppConstants::CONFIG_MAX_RESPONSE_REL_COUNT);
            $val = max(1, $configVal);
        }

        if(array_key_exists(AppConstants::REQUEST_MAX_RESPONSE_REL_COUNT, $_GET)){
            $reqVal = $_GET[AppConstants::REQUEST_MAX_RESPONSE_REL_COUNT];
            $val = max(1, $reqVal);
        }

        return $val;
    }
}
