<?php

namespace App\Service;

use ApiPlatform\Core\Api\IriConverterInterface;
use App\Constants\AppConstants;
use App\Constants\AttributeLevelOperators;
use App\Constants\CacheConstants;
use App\Constants\EntityFieldConstants;
use App\Doctrine\ORM\FilterQueryBuilder;
use App\Entity\Base\BaseEntity;
use App\Entity\Query\Query;
use App\Entity\Rules\PreprocessingRule;
use App\Traits\Trackable;
use DateTime;
use Doctrine\Inflector\InflectorFactory;
use Doctrine\ORM\EntityManager;
use Doctrine\ORM\EntityManagerInterface;
use Doctrine\ORM\Mapping\ClassMetadataInfo;
use Psr\Log\LoggerInterface;
use Symfony\Component\DependencyInjection\ContainerInterface;
use Symfony\Component\HttpFoundation\Request;
use Symfony\Contracts\Cache\TagAwareCacheInterface;

class PreprocessingRuleService
{
    private \Doctrine\Inflector\Inflector $inflector;

    public function __construct(
        protected ContainerInterface     $container,
        protected FilterQueryBuilder     $filterQueryBuilder,
        protected TagAwareCacheInterface $cache,
        protected MetadataHelper         $metadataHelper,
        protected IriConverterInterface  $iriConverter,
        protected EntityManagerInterface $entityManager,
        protected LoggerInterface        $logger
    ) {
        $this->inflector = InflectorFactory::create()->build();
    }


    /**
     * @throws \Exception
     */
    public function getRules($entityType)
    {
        $cacheKey = CacheConstants::PREPROCESSING_RULES . "_" .  strtoupper($entityType);
        $item = $this->cache->getItem($cacheKey);
        if (!$item->isHit()) {
            /**
             * @var PreprocessingRule[] $preprocessingRules
             */
            $preprocessingRules = [];
            $ruleRecords = $this->findActiveRules($entityType);
            foreach ($ruleRecords as $rule) {
                $preprocessingRules[] = $this->prepareRuleCacheObject($rule);
            }
            $this->updateRulesCache($item, $preprocessingRules);
        }
        return $item->get();
    }

    /**
     * @param $item
     * @param $data
     * @return void
     */
    private function updateRulesCache($item, $data): void
    {
        $item->set($data);
        $item->expiresAfter(AppConstants::SECONDS_IN_ONE_DAY);
        $item->tag(CacheConstants::PREPROCESSING_RULES);
        $this->cache->save($item);
    }


    /**
     * @return PreprocessingRule[]
     */
    private function findActiveRules($entityType): array
    {
        $preprocessingRules = [];
        $preprocessingRuleRepo = $this->entityManager->getRepository(PreprocessingRule::class);
        $preprocessingRulesRecords = $preprocessingRuleRepo->findBy(
            ["isActive" => true, "entityType" => $entityType],
            array('priority' => 'ASC', 'createDate' => 'ASC')
        );

        foreach ($preprocessingRulesRecords as $key => $value) {
            $preprocessingRules[] = $value;
        }

        return $preprocessingRules;
    }


    /**
     * @param string $entity
     * @param mixed $object
     * @param string $requestMethod
     * @return bool
     */
    public function validateRecord(string $entity, mixed $object): bool
    {
        return in_array($entity, AppConstants::SYSTEM_ENTITY) || !$object instanceof BaseEntity;
    }

    /**
     * @return int
     */
    public function getMaximumDateFilter(): int
    {
        $maxDays = AppConstants::RULE_MAX_DAYS_FILTER_DEFAULT_VALUE;
        if ($this->container->hasParameter(AppConstants::RULE_MAX_DAYS_FILTER_KEY)) {
            $maxDays = $this->container->getParameter(AppConstants::RULE_MAX_DAYS_FILTER_KEY);
        }
        return min($maxDays, 30);
    }

    /**
     * @param mixed $value
     * @param mixed $data
     * @param bool $postSave
     * @return array|mixed
     */
    public function prepareRuleCriteria(mixed $value, mixed $data, bool $postSave): mixed
    {
        $criteria = $value[EntityFieldConstants::RULE_ATTR_CRITERIA];
        $noOfDays = $criteria['days'] ?? 7;
        $finalCondition = $this->prepareDateCriteria($noOfDays);

        $condition = $criteria['condition'];
        if (isset($condition[Query::KEYWORD_FILTERS])) {
            unset($condition[Query::KEYWORD_SORT]);
            unset($condition[Query::KEYWORD_LIMIT]);
            $this->resolveFilterCriteria($data, $condition);
            $finalCondition[Query::KEYWORD_FILTERS][] = $condition;
        }

        // add default sort of createDate
        $finalCondition['sort'] = array(["field" =>  "createDate", "direction" => "ASC"]);
        return $finalCondition;
    }

    /**
     * @param array $filter
     * @param $data
     * @return void
     */
    private function resolveFilterCriteria($data, &$filter): void
    {
        if ($filter && array_key_exists(Query::KEYWORD_LOGIC, $filter) && array_key_exists(Query::KEYWORD_FILTERS, $filter)) {
            $nestedFilters = &$filter[Query::KEYWORD_FILTERS];
            foreach ($nestedFilters as &$nestedFilter) {
                $this->resolveFilterCriteria($data, $nestedFilter);
            }
        } elseif ($filter && array_key_exists('field', $filter)) {
            $operator = strtolower($filter['operator']);
            if ("exists" == strtolower($operator)) {
                $field = $filter['field'];
                $filter['operator'] = AttributeLevelOperators::EQUALS;
                $filter['value'] = $data[$field];
                $filter['type'] = 'string';
            }
        }
    }

    private function prepareDateCriteria(int $days): array
    {
        $defaultDays = $this->getMaximumDateFilter();
        if ($days > $defaultDays) $days = $defaultDays;

        return [
            Query::KEYWORD_SORT => [],
            Query::KEYWORD_LIMIT => 30,
            Query::KEYWORD_LOGIC => "AND",
            Query::KEYWORD_FILTERS => [[
                AppConstants::QUERY_FILTER_TYPE => AppConstants::FILTER_TYPE_PRIMITIVE,
                AppConstants::QUERY_FILTER_FIELD => "createDate",
                AppConstants::QUERY_FILTER_VAL => $this->getDateDiff($days),
                AppConstants::QUERY_FILTER_OPERATOR => AttributeLevelOperators::GREATER_THAN_EQUALS,
                AppConstants::QUERY_FILTER__OPERATOR => AttributeLevelOperators::GREATER_THAN_EQUALS
            ]]
        ];
    }

    private function getDateDiff(int $days): string
    {
        $date = date('Y-m-d');
        return date('Y-m-d', strtotime($date . ' - ' . $days . ' days'));
    }

    /**
     * @param PreprocessingRule $value
     * @return array
     */
    public function prepareRuleCacheObject(PreprocessingRule $value): array
    {
        return [
            EntityFieldConstants::RULE_ATTR_APPLICATION_ON => $value->getApplicableOn(),
            EntityFieldConstants::RULE_ATTR_SKIP_PL_EXEC => $value->getSkipPlaybookExecution(),
            EntityFieldConstants::RULE_ATTR_ACTIVE => $value->getIsActive(),
            EntityFieldConstants::RULE_ATTR_PRIORITY => $value->getPriority(),
            EntityFieldConstants::RULE_ATTR_ENTITY_TYPE => $value->getEntityType(),
            EntityFieldConstants::RULE_ATTR_ACTION => $value->getAction(),
            EntityFieldConstants::RULE_ATTR_CRITERIA => $value->getCriteria(),
            EntityFieldConstants::RULE_ATTR_ACTION_TYPE => $value->getActionType()->getItemValue(),
            EntityFieldConstants::RULE_ATTR_NAME => $value->getName(),
            EntityFieldConstants::RULE_ATTR_UUID => $value->getUuid(),
            EntityFieldConstants::RULE_ATTR_END_DATE => $value->getEndDate()
        ];
    }

    /**
     * @param array $rules
     * @param $recordType ["incoming" or "existing"]
     * @return array
     */
    public function filterByRecordType(array $rules, $recordType): array
    {
        return array_filter($rules, function ($value) use ($recordType) {
            return strcasecmp($value[EntityFieldConstants::RULE_ATTR_APPLICATION_ON], $recordType) == 0;
        });
    }

    /**
     * @param array $rules
     * @param $ruleType
     * @return array
     */
    public function filterByRuleType(array $rules, $ruleType): array
    {
        return array_filter($rules, function ($value) use ($ruleType) {
            return strcasecmp($value[EntityFieldConstants::RULE_ATTR_ACTION_TYPE], $ruleType) == 0;
        });
    }

    /**
     * @param array $rules
     * @return array
     */
    public function filterByEndDate(array $rules): array
    {
        return array_filter($rules, function ($value) {
            $currenTime = new DateTime();
            $endDate = $value[EntityFieldConstants::RULE_ATTR_END_DATE];
            return is_null($endDate) || ($endDate >= $currenTime);
        });
    }

    /**
     * @param array $rules
     * @return array
     */
    public function sortByPriority(array $rules): array
    {
        usort($rules, function ($a, $b) {
            return strcmp($b[EntityFieldConstants::RULE_ATTR_PRIORITY], $a[EntityFieldConstants::RULE_ATTR_PRIORITY]);
        });
        return $rules;
    }

    /**
     * @param array $rules
     * @param mixed $entityType
     * @return array
     */
    public function filterByEntityType(array $rules, mixed $entityType): array
    {
        return array_filter($rules, function ($value) use ($entityType) {
            return strcasecmp($value[EntityFieldConstants::RULE_ATTR_ENTITY_TYPE], $entityType) == 0;
        });
    }

    /**
     * @param BaseEntity $object
     * @return bool
     */
    public function isModuleTrackable(BaseEntity $object): bool
    {
        return in_array(Trackable::class, class_uses(get_class($object)));
    }


    /**
     * @param $entityClass
     * @param $query
     * @param $preprocess
     * @return float|int|mixed|string
     */
    public function executeRuleQuery($entityClass, $query, $preprocess): mixed
    {
        $returnFields = [];
        if ($preprocess) $returnFields = ["o.uuid"];

        /** @var EntityManager $manager */
        $repository = $this->entityManager->getRepository($entityClass);
        $queryBuilder = $repository->createQueryBuilder('o');

        if (!(isset($query['logic']) && isset($query['filters']))) {
            $query['logic'] = "AND";
            $query['filters'] = [];
        }

        $this->filterQueryBuilder->fromQueryObject($query, $queryBuilder, $entityClass, $returnFields, !$preprocess);
        return $queryBuilder->setMaxResults(1)->getQuery()->execute();
    }

    /**
     * @param $classMetadata
     * @param $queryBuilder
     * @return void
     */
    public function entity_association($classMetadata, $queryBuilder): void
    {
        // $classMetadata = $entityManager->getClassMetadata($entityClass);
        // $this->entity_association($classMetadata, $queryBuilder);

        foreach ($classMetadata->getAssociationNames() as $i => $association) {
            $mapping = $classMetadata->associationMappings[$association];

            if (ClassMetadataInfo::FETCH_EAGER === $mapping['fetch']) {
                $queryBuilder->leftJoin('o.' . $association, 'a' . $i);
                $queryBuilder->addSelect('a' . $i);
            }
        }
    }

    /**
     * @param $rule
     * @return array
     */
    public function prepareRuleExecutionFailedBroadcastData($rule): array
    {
        $auditData = [];
        $auditData['rule'] = $rule;

        return [
            AppConstants::COMPONENT => AppConstants::CRUD_HUB,
            AppConstants::ENTITY_TYPE =>  AppConstants::ENTITY_PREPROCESSING_RULES,
            AppConstants::HTTP_METHOD => AppConstants::AUDIT_OPERATION_RULE_FAILED,
            AppConstants::REMOTE_ADDR => $_SERVER['HTTP_X_FORWARDED_FOR'] ?? $_SERVER['REMOTE_ADDR'],
            AppConstants::TRANSACTION_DATE => (int)($_SERVER['REQUEST_TIME_FLOAT'] * 1000),
            AppConstants::DELTA_DATA => $auditData,
        ];
    }
}
