<?php
/* Copyright start
  Copyright (C) 2008 - 2024 Fortinet Inc.
  All rights reserved.
  FORTINET CONFIDENTIAL & FORTINET PROPRIETARY SOURCE CODE
  Copyright end */

namespace App\Service;

use Psr\Log\LoggerInterface;
use App\Constants\AppConstants;
use App\Constants\CacheConstants;
use GuzzleHttp\Psr7\Request as GuzzleRequest;
use Symfony\Contracts\Cache\TagAwareCacheInterface;
use Symfony\Component\DependencyInjection\ContainerInterface;
use Symfony\Component\HttpKernel\Exception\BadRequestHttpException;

class SKUService
{
    protected $dasClient;
    protected $dasUri;
    private $headers = array("Content-type" => "application/json", "X-ADMIN-USER" => true);

    public function __construct(
        private LoggerInterface $logger,
        protected ContainerInterface $containers,
        protected TagAwareCacheInterface $cache
    ) {
        $this->dasClient = $this->containers->get("app.proxy.client.das");
        $this->dasUri = $this->containers->getParameter('das.uri');
    }

    public function getLicenseDetailsFromDas()
    {
        $timEnabled = false;
        $isDistributedInstance = false;

        $instanceRole = AppConstants::MASTER_ROLE_NAME;
        $supportedIngestionLimit = AppConstants::TOTAL_INGESTION_FEEDS_LIMIT;
        $supportedQueryLimit = AppConstants::TOTAL_RESTRICTED_FEEDS_COUNT;

        $dasRequests = new GuzzleRequest('get', $this->dasUri . '/license/?param=license_details', $this->headers);
        try {
            $dasResponse = $this->dasClient->send($dasRequests);
            $dasResponseData = $dasResponse->getBody()->getContents();
            $jsonDecodeResponse = json_decode($dasResponseData, true);
            $resCode = $dasResponse->getStatusCode();
            if ($resCode < 300) {
                $isDistributedInstance = $jsonDecodeResponse['details']['is_distributed'] ?? false;
                $instanceRole = $isDistributedInstance ? $jsonDecodeResponse['details']['role'][0] : null;

                // Contracts key is always expected
                $timEnabled = $jsonDecodeResponse['details']['contracts']['has_tim_subscription'] ?? false;
                $supportedIngestionLimit = $jsonDecodeResponse['details']['contracts']['tim_supported_ingestion_limit'] ?? AppConstants::TOTAL_INGESTION_FEEDS_LIMIT;
                $supportedQueryLimit = $jsonDecodeResponse['details']['contracts']['tim_supported_query_limit'] ?? AppConstants::TOTAL_RESTRICTED_FEEDS_COUNT;

                $this->setCache($isDistributedInstance, $instanceRole, $timEnabled, $supportedIngestionLimit, $supportedQueryLimit);
                return [$isDistributedInstance, $instanceRole, $timEnabled, $supportedIngestionLimit, $supportedQueryLimit];
            }
        } catch (BadRequestHttpException | \Exception $error) {
            $this->logger->info($error->getMessage());
        }
        return [$isDistributedInstance, $instanceRole, $timEnabled, $supportedIngestionLimit, $supportedQueryLimit];
    }

    public function getHAClusterNodesFromDas()
    {
        $nodeList = array();
        $retryCount = 5;
        while ($retryCount >0 ){
            try {
                $cachedNodes = $this->cache->getItem("haNodes");
                if (!$cachedNodes->isHit()) {
                    $dasRequests = new GuzzleRequest('get', $this->dasUri . '/cluster/', ["Content-type" => "application/json"]);
                    $dasResponse = $this->dasClient->send($dasRequests);
                    $resCode = $dasResponse->getStatusCode();
                    if ($resCode != 200) {
                        error_log('Failed to fetch ha nodes from das');
                    }
                    $dasResponseData = $dasResponse->getBody()->getContents();
                    $jsonDecodeResponse = json_decode($dasResponseData, true);
                    $fortisoarNodes = $jsonDecodeResponse["nodes"];
                    foreach ($fortisoarNodes as $node){
                        if ($node["status"] == "active"){
                            $nodeList[] = $node["nodeName"];
                        }
                    }
                    if( !empty($nodeList) ){
                        $cachedNodes->set($nodeList);
                        $cachedNodes->expiresAfter(7200);
                        $cachedNodes->tag(CacheConstants::HA_NODES);
                        $this->cache->save($cachedNodes);
                    }
                    else{
                        throw new \Exception("HA cluster node list cannot be empty");
                    }
                    
                }
                $nodeList = $cachedNodes->get();
                break;
            } catch (\Exception $error) {
                error_log('cyops-auth service request failed');
                error_log($error->getMessage());
                sleep(2);
                $retryCount -=1;
            }
        }

        return $nodeList;
    }

    private function setCache($isDistributedInstance, $instanceRole, $timEnabled, $supportedIngestionLimit, $supportedQueryLimit)
    {
        $item = $this->cache->getItem(CacheConstants::IS_DISTRIBUTED_INSTANCE);
        $item->set($isDistributedInstance);
        $item->expiresAfter(AppConstants::CACHE_EXPIRY_FOR_LICENSE);
        $item->tag(CacheConstants::MSSP_TYPE_ROLE);
        $this->cache->save($item);

        $item = $this->cache->getItem(CacheConstants::INSTANCE_ROLE);
        $item->set($instanceRole);
        $item->expiresAfter(AppConstants::CACHE_EXPIRY_FOR_LICENSE);
        $item->tag(CacheConstants::MSSP_TYPE_ROLE);
        $this->cache->save($item);

        $item = $this->cache->getItem(CacheConstants::IS_TIM_ENABLED);
        $item->set($timEnabled);
        $item->expiresAfter(AppConstants::CACHE_EXPIRY_FOR_LICENSE);
        $item->tag(CacheConstants::IS_TIM_ENABLED);
        $this->cache->save($item);

        $item = $this->cache->getItem(CacheConstants::TIM_SUPPORTED_INGESTION_LIMIT);
        $item->set($supportedIngestionLimit);
        $item->expiresAfter(AppConstants::CACHE_EXPIRY_FOR_LICENSE);
        $item->tag(CacheConstants::TIM_SUPPORTED_INGESTION_LIMIT);
        $this->cache->save($item);

        $item = $this->cache->getItem(CacheConstants::TIM_SUPPORTED_QUERY_LIMIT);
        $item->set($supportedQueryLimit);
        $item->expiresAfter(AppConstants::CACHE_EXPIRY_FOR_LICENSE);
        $item->tag(CacheConstants::TIM_SUPPORTED_QUERY_LIMIT);
        $this->cache->save($item);
    }
}
